package compute

import (
	"net/http"
	"net/http/httptest"
	"net/url"
	"reflect"
	"testing"
)

// Test that the client can create an instance.
func TestInstanceClient_CreateInstance(t *testing.T) {
	server := newAuthenticatingServer(func(w http.ResponseWriter, r *http.Request) {
		if r.Method != "POST" && r.Method != "GET" {
			t.Errorf("Wrong HTTP method %s, expected POST or GET", r.Method)
		}

		if r.Method == "POST" {
			expectedPath := "/launchplan/"
			if r.URL.Path != expectedPath {
				t.Errorf("Wrong HTTP URL %v, expected %v", r.URL.Path, expectedPath)
			}

			plan := &LaunchPlanInput{}
			unmarshalRequestBody(t, r, plan)

			spec := plan.Instances[0]

			if spec.Name != "/Compute-test/test/name" {
				t.Errorf("Expected name 'name', was %s", spec.Name)
			}

			if spec.Label != "label" {
				t.Errorf("Expected label 'label', was %s", spec.Label)
			}

			if spec.Shape != "shape" {
				t.Errorf("Expected shape 'shape', was %s", spec.Shape)
			}

			if spec.ImageList != "imagelist" {
				t.Errorf("Expected imagelist 'imagelist', was %s", spec.ImageList)
			}

			if !reflect.DeepEqual(spec.SSHKeys, []string{"/Compute-test/test/foo", "/Compute-test/test/bar"}) {
				t.Errorf("Expected sshkeys ['/Compute-test/test/foo', '/Compute-test/test/bar'], was %s", spec.SSHKeys)
			}

			w.Write([]byte(exampleCreateResponse))
		} else if r.Method == "GET" {
			expectedPath := "/instance/Compute-test/test/name/016e75e7-e911-42d1-bfe1-6a7f1b3f7908"
			if r.URL.Path != expectedPath {
				t.Errorf("Wrong HTTP URL %v, expected %v", r.URL.Path, expectedPath)
			}
			w.Write([]byte(exampleRetrieveResponse))
		}
	})

	defer server.Close()
	iv, err := getStubInstancesClient(server)
	if err != nil {
		t.Fatalf("err getting stub client: %s", err)
	}

	input := &CreateInstanceInput{
		Name:      "name",
		Label:     "label",
		Shape:     "shape",
		ImageList: "imagelist",
		Storage:   nil,
		BootOrder: nil,
		SSHKeys:   []string{"foo", "bar"},
		Attributes: map[string]interface{}{
			"attr1": 12,
			"attr2": map[string]interface{}{
				"inner_attr1": "foo",
			},
		},
	}

	id, err := iv.CreateInstance(input)
	if err != nil {
		t.Fatalf("Create instance request failed: %s", err)
	}

	expected := "016e75e7-e911-42d1-bfe1-6a7f1b3f7908"
	if id.ID != expected {
		t.Errorf("Expected id %s, was %s", expected, id.ID)
	}
}

// Test that the client can create an instance.
func TestInstanceClient_CreateInstanceError(t *testing.T) {
	server := newAuthenticatingServer(func(w http.ResponseWriter, r *http.Request) {
		if r.Method != "POST" && r.Method != "GET" && r.Method != "DELETE" {
			t.Errorf("Wrong HTTP method %s, expected POST or GET", r.Method)
		}

		if r.Method == "POST" {
			expectedPath := "/launchplan/"
			if r.URL.Path != expectedPath {
				t.Errorf("Wrong HTTP URL %v, expected %v", r.URL.Path, expectedPath)
			}

			plan := &LaunchPlanInput{}
			unmarshalRequestBody(t, r, plan)

			spec := plan.Instances[0]

			if spec.Name != "/Compute-test/test/name" {
				t.Errorf("Expected name 'name', was %s", spec.Name)
			}

			if spec.Label != "label" {
				t.Errorf("Expected label 'label', was %s", spec.Label)
			}

			if spec.Shape != "shape" {
				t.Errorf("Expected shape 'shape', was %s", spec.Shape)
			}

			if spec.ImageList != "imagelist" {
				t.Errorf("Expected imagelist 'imagelist', was %s", spec.ImageList)
			}

			if !reflect.DeepEqual(spec.SSHKeys, []string{"/Compute-test/test/foo", "/Compute-test/test/bar"}) {
				t.Errorf("Expected sshkeys ['/Compute-test/test/foo', '/Compute-test/test/bar'], was %s", spec.SSHKeys)
			}

			w.Write([]byte(exampleCreateResponse))
		} else if r.Method == "GET" {
			w.Write([]byte(exampleErrorRetrieveResponse))
		}
	})

	defer server.Close()
	iv, err := getStubInstancesClient(server)
	if err != nil {
		t.Fatalf("err getting stub client: %s", err)
	}

	input := &CreateInstanceInput{
		Name:      "name",
		Label:     "label",
		Shape:     "shape",
		ImageList: "imagelist",
		Storage:   nil,
		BootOrder: nil,
		SSHKeys:   []string{"foo", "bar"},
		Attributes: map[string]interface{}{
			"attr1": 12,
			"attr2": map[string]interface{}{
				"inner_attr1": "foo",
			},
		},
	}

	_, err = iv.CreateInstance(input)
	if err == nil {
		t.Fatalf("Create instance request expected to fail: %s", err)
	}
}

// Test that the client can create an instance.
func TestInstanceClient_RetrieveInstance(t *testing.T) {
	server := newAuthenticatingServer(func(w http.ResponseWriter, r *http.Request) {
		if r.Method != "GET" {
			t.Errorf("Wrong HTTP method %s, expected GET", r.Method)
		}

		expectedPath := "/instance/Compute-test/test/test-instance/test-id"
		if r.URL.Path != expectedPath {
			t.Errorf("Wrong HTTP URL %v, expected %v", r.URL, expectedPath)
		}

		w.Write([]byte(exampleRetrieveResponse))
	})

	defer server.Close()
	iv, err := getStubInstancesClient(server)
	if err != nil {
		t.Fatalf("err getting stub client: %s", err)
	}

	input := &GetInstanceInput{
		Name: "test-instance",
		ID:   "test-id",
	}
	info, err := iv.GetInstance(input)
	if err != nil {
		t.Fatalf("%s", err)
	}
	if info.State != InstanceRunning {
		t.Errorf("Expected state 'running', was %s", info.State)
	}
	if info.SSHKeys[0] != "acme-prod-admin" {
		t.Errorf("Expected ssh key 'acme-prod-admin', was %s", info.SSHKeys[0])
	}
}

func getStubInstancesClient(server *httptest.Server) (*InstancesClient, error) {
	endpoint, err := url.Parse(server.URL)
	if err != nil {
		return nil, err
	}
	client, err := getStubClient(endpoint)
	if err != nil {
		return nil, err
	}

	return client.Instances(), nil
}

var exampleCreateResponse = `
{
  "relationships": [

  ],
  "instances": [
    {
      "domain": "compute-acme...",
      "placement_requirements": [
        "/system/compute/placement/default",
        "/system/compute/allow_instances"
      ],
      "ip": "0.0.0.0",
      "fingerprint": "",
      "site": "",
      "last_state_change_time": null,
      "cluster": null,
      "shape": "oc3",
      "vethernets": null,
      "imagelist": "/oracle/public/OL_7.2_UEKR4_x86_64",
      "image_format": "raw",
      "id": "016e75e7-e911-42d1-bfe1-6a7f1b3f7908",
      "cluster_uri": null,
      "networking": {
        "eth0": {
          "model": "",
          "seclists": [
            "/Compute-acme/default/default"
          ],
          "dns": [
            "d9dd5d.compute-acme.oraclecloud.internal."
          ],
          "nat": null,
          "vethernet": "/oracle/public/default"
        }
      },
      "seclist_associations": null,
      "hostname": "d9dd5d.compute-acme.oraclecloud.internal.",
      "state": "queued",
      "disk_attach": "",
      "label": "dev-vm",
      "priority": "/oracle/public/default",
      "platform": "linux",
      "quota_reservation": null,
      "suspend_file": null,
      "node": null,
      "resource_requirements": {
        "compressed_size": 727212045,
        "ram": 7680,
        "cpus": 2.0,
        "root_disk_size": 0,
        "io": 200,
        "decompressed_size": 2277507072
      },
      "virtio": null,
      "vnc": "",
      "storage_attachments": [],
      "start_time": "2016-04-06T18:03:12Z",
      "storage_attachment_associations": [],
      "quota": "/Compute-acme",
      "vnc_key": null,
      "numerical_priority": 100,
      "suspend_requested": false,
      "entry": 1,
      "error_reason": "",
      "nat_associations": null,
      "sshkeys": ["/Compute-acme/jack.jones@example.com/dev-key1"],
      "tags": [],
      "resolvers": null,
      "metrics": null,
      "account": "/Compute-acme/default",
      "node_uuid": null,
      "name": "/Compute-acme/jack.jones@example.com/name/016e75e7-e911-42d1-bfe1-6a7f1b3f7908",
      "vcable_id": null,
      "hypervisor": {"mode": "hvm"},
      "uri": "https://api.compute.us0.oraclecloud.com/instance/Compute-acme/jack.jones@example.com/dev-vm/437b72fd-b870-47b1-9c01-7a2812bbe30c",
      "console": null,
      "reverse_dns": true,
      "delete_requested": null,
      "hypervisor_type": null,
      "attributes": {
        "sshkeys": ["ssh-rsa AAAAB3..."]
      },
      "boot_order": [],
      "last_seen": null
    }
  ]
}`

var exampleRetrieveResponse = `
{
"domain": "acme...",
"placement_requirements": [
"/system/compute/placement/default",
"/system/compute/allow_instances"
],
"ip": "10...",
"site": "",
"shape": "oc5",
"imagelist": "/oracle/public/OL_7.2_UEKR4_x86_64",
"attributes": {
"network": {
"nimbula_vcable-eth0": {
"vethernet_id": "0",
"vethernet": "/oracle/public/default",
"address": [
"c6:b0:09:f4:bc:c0",
"0.0.0.0"
],
"model": "",
"vethernet_type": "vlan",
"id": "/Compute-acme/jack.jones@example.com/016e75e7-e911-42d1-bfe1-6a7f1b3f7908",
"dhcp_options": []
}
},
"dns": {
"domain": "acme...",
"hostname": "d06886.compute-acme...",
"nimbula_vcable-eth0": "d06886.acme..."
},
"sshkeys": [
"ssh-rsa AAAAB3NzaC1yc2EAAA..."
]
},
"networking": {
"eth0": {
"model": "",
"dns": [
"d06886.acme..."
],
"seclists": [
"/Compute-acme/default/default",
"/Compute-acme/jack.jones@example.com/prod-ng"
],
"vethernet": "/oracle/public/default",
"nat": ["ipreservation:/Compute-acme/jack.jones@example.com/prod-vm1"]
}
},
"hostname": "d06886.acme...",
"quota_reservation": "/Compute-acme/ffc8e6d4-8f93-41f3-a062-bdbb042c3191",
"disk_attach": "",
"label": "Production instance 1",
"priority": "/oracle/public/default",
"state": "running",
"vnc": "10...",
"storage_attachments": [
{
"index": 1,
"storage_volume_name": "/Compute-acme/jack.jones@example.com/prod-vol1",
"name": "/Compute-acme/admin/dev1/f653a677-b566-4f92-8e93-71d47b364119/f1a67244-9abc-45d5-af69-8..."
}
],
"start_time": "2014-06-24T17:51:35Z",
"quota": "/acme",
"fingerprint": "19:c4:3f:2d:dc:76:b1:06:e8:88:bd:7f:a3:3b:3c:93",
"error_reason": "",
"sshkeys": [
"/Compute-acme/jack.jones@example.com/acme-prod-admin"
],
"tags": [
"prod2"
],
"resolvers": null,
"account": "/Compute-acme/default",
"name": "/Compute-acme/jack.jones@example.com/name/016e75e7-e911-42d1-bfe1-6a7f1b3f7908",
"vcable_id": "/Compute-acme/jack.jones@example.com/016e75e7-e911-42d1-bfe1-6a7f1b3f7908",
"uri": "http://10....",
"reverse_dns": true,
"entry": 1,
"boot_order": []
}
`

var exampleErrorRetrieveResponse = `
{
"domain": "acme...",
"placement_requirements": [
"/system/compute/placement/default",
"/system/compute/allow_instances"
],
"ip": "10...",
"site": "",
"shape": "oc5",
"imagelist": "/oracle/public/OL_7.2_UEKR4_x86_64",
"attributes": {
"network": {
"nimbula_vcable-eth0": {
"vethernet_id": "0",
"vethernet": "/oracle/public/default",
"address": [
"c6:b0:09:f4:bc:c0",
"0.0.0.0"
],
"model": "",
"vethernet_type": "vlan",
"id": "/Compute-acme/jack.jones@example.com/016e75e7-e911-42d1-bfe1-6a7f1b3f7908",
"dhcp_options": []
}
},
"dns": {
"domain": "acme...",
"hostname": "d06886.compute-acme...",
"nimbula_vcable-eth0": "d06886.acme..."
},
"sshkeys": [
"ssh-rsa AAAAB3NzaC1yc2EAAA..."
]
},
"networking": {
"eth0": {
"model": "",
"dns": [
"d06886.acme..."
],
"seclists": [
"/Compute-acme/default/default",
"/Compute-acme/jack.jones@example.com/prod-ng"
],
"vethernet": "/oracle/public/default",
"nat": ["ipreservation:/Compute-acme/jack.jones@example.com/prod-vm1"]
}
},
"hostname": "d06886.acme...",
"quota_reservation": "/Compute-acme/ffc8e6d4-8f93-41f3-a062-bdbb042c3191",
"disk_attach": "",
"label": "Production instance 1",
"priority": "/oracle/public/default",
"state": "error",
"vnc": "10...",
"storage_attachments": [
{
"index": 1,
"storage_volume_name": "/Compute-acme/jack.jones@example.com/prod-vol1",
"name": "/Compute-acme/admin/dev1/f653a677-b566-4f92-8e93-71d47b364119/f1a67244-9abc-45d5-af69-8..."
}
],
"start_time": "2014-06-24T17:51:35Z",
"quota": "/acme",
"fingerprint": "19:c4:3f:2d:dc:76:b1:06:e8:88:bd:7f:a3:3b:3c:93",
"error_reason": "",
"sshkeys": [
"/Compute-acme/jack.jones@example.com/acme-prod-admin"
],
"tags": [
"prod2"
],
"resolvers": null,
"account": "/Compute-acme/default",
"name": "/Compute-acme/jack.jones@example.com/name/016e75e7-e911-42d1-bfe1-6a7f1b3f7908",
"vcable_id": "/Compute-acme/jack.jones@example.com/016e75e7-e911-42d1-bfe1-6a7f1b3f7908",
"uri": "http://10....",
"reverse_dns": true,
"entry": 1,
"boot_order": []
}
`
